/* Complicated version that minimized allocation while not using
   the original array to store intermediate values (which would be
   useful if intermediate values were something other than ints). */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "popp_thread.h"

static int *array, *converted;
static int length = 100000;
static int base_accum = 0;

static void make_array()
{
  int i;

  array = (int *)malloc(length * sizeof(int));
  for (i = 0; i < length; i++) {
    array[i] = 1;
  }
  array[0] = 2; /* results: 2 300002 599996 */
  /* array[0] = 3; => results: 6 300006 600000 */
}

static int convert(int v)
{
  int i;

  /* Arbitrary time-consuming calculation: */
  for (i = 0; i < 1000; i++) {
    v = ((v * v) + v) % 10;
  }

  return v;
}

static int deconvert(int v)
{
  return convert(v);
}

static int do_section(int start, int end, int accum)
{
  int i;

  for (i = start; i < end; i++) {
    accum += convert(array[i]);
    array[i] = deconvert(accum);
  }

  return accum;
}

/*****************************************/
/* No-threads case:                      */

int sequential_sums()
{
  return do_section(0, length, base_accum);
}

/*****************************************/
/* Two-threads case:                    */

static int section_2_start;
static int section_3_start;

static int first_total;
static int second_total;
static int third_total;

void do_first_section(int id)
{
  first_total = do_section(0, section_2_start, base_accum);
}

static int convert_section(int start, int end)
{
  int i, v, accum = 0;

  for (i = start; i < end; i++) {
    v = convert(array[i]);
    converted[i-section_2_start] = v;
    accum += v;
  }

  return accum;
}

void convert_second_section(int id)
{
  second_total = convert_section(section_2_start, section_3_start);
  third_total = convert_section(section_3_start, length);
}

static int do_converted_section(int start, int end, int accum)
{
  int i;

  for (i = start; i < end; i++) {
    accum += converted[i-section_2_start];
    array[i] = deconvert(accum);
  }

  return accum;
}

void do_second_section(int id)
{
  do_converted_section(section_2_start, section_3_start, first_total);
}

void do_third_section(int id)
{
  do_converted_section(section_3_start, length, first_total + second_total);
}

int parallel_sums()
{
  section_2_start = length / 3; /* <- change 3 to 2 if deconvert() is cheap */
  section_3_start = section_2_start + ((length - section_2_start) / 2);

  converted = (int *)malloc((length - section_2_start) * sizeof(int));

  thread_create(do_first_section, 0);
  thread_create(convert_second_section, 0);
  join_threads();
  thread_create(do_second_section, 0);
  thread_create(do_third_section, 0);
  join_threads();

  free(converted);

  return first_total + second_total + third_total;
}

/*****************************************/

void multi_sums()
{
  int pieces = 10, i, a;
  int *save_array, save_length, piece_length;

  save_array = array;
  save_length = length;

  piece_length = length / pieces;

  for (i = 0; i < pieces; i++) {
    array = save_array + (piece_length * i);
    length = piece_length;
    a = parallel_sums(); /* could use sequential_sums() here */
    base_accum = a;
  }
  
  array = save_array;
  length = save_length;
}

/*****************************************/

int main(int argc, char **argv) 
{
  make_array();
  
  sequential_sums();

  printf("%d %d %d\n", array[0], array[length / 2], array[length - 1]);

  return 0;
}
